#include <stdio.h>
#include <dos.h>
#include <math.h>
#include <devices.h>

typedef union {
	word	Word;
	int	Int;
	byte	Byte [2];
	char	Char [2];
} reg16;

typedef union {
	dword	Dword;
	long 	Long;
	word	Word [2];
	int	Int [2];
	byte	Byte [4];
	char	Char [4];
} reg32;

const word /* UART registers offset */
	THR 	= 0,			/* Transmit Holding Register, DLAB = 0 */
	RBR	= 0,			/* Receive Buffer Register, DLAB = 0 */
	DLL 	= 0,			/* Divisor Latch LSB, DLAB = 1 */
	DLM	= 1,			/* Divisor Latch MSB, DLAB = 1 */
	IER 	= 1,			/* Interrupt Enable Register, DLAB = 0 */
	LCR 	= 3,			/* Line Control Register */
	MCR 	= 4,			/* Modem Control Register */
	LSR 	= 5;			/* Line Status Register */

const byte /* LCR bit masks */
	WLEN 	= 0x03,		/* Word Length Select Bits 0 & 1 */
	STB	= 0x04,		/* Number of Stop Bits */
	PEN 	= 0x08,		/* Parity Enable */
	EVEN	= 0x10,		/* Even Parity Select */
	STICK	= 0x20,		/* Stick Parity */
	DLAB 	= 0x80;		/* Divisor Latch Access Bit */

const byte	/* LSR bit masks */
	DR		= 0x01,		/* Data Ready */
	OR		= 0x02,		/* Overrun Error */
	PE		= 0x04,		/* Parity Error */
	FE		= 0x08,		/* Framing Error */
	BI		= 0x10,		/* Break Interrupt */
	THRE	= 0x20,		/* Transmitter Holding Register Empty */
	TSRE	= 0x40;		/* Transmitter Shift Register Empty */

const byte /* MCR bit masks */
	DTR	= 0x01,		/* Data Terminal Ready */
	LOOP	= 0x10;		/* Loop back test */

const byte /* Codes sent from Host to Device */
	RTR 	= 0xD0,			/* ready to read */
	EXTD 	= 0xFF;			/* Extended Parameter or Command */

const byte /* Interrupt enable bits */
	ERLS = 0x040;

/* Codes sent from Device to Host */
#define	ACK	0x00			/* acknowledge */
#define	MSCD	0x3C			/* error code follows */
#define	BUSY	0x3D			/* device busy */
#define	NACK	0x3E			/* not acknowledge */
#define	MSGS	0x3F			/* message string follows */

const byte /* Commands recognized by the Device */
	READ_DATA	= 1,   /* old 0,1 */
	WRITE_DATA	= 2,
	READ_EE		= 0,  /* old 2,3 */
	WRITE_EE		= 0,
	INVOKE_CMD	= 4;

const byte /* RS485 related constants */
	RS485_REG	= 7,
	RS485_OUT	= 2,
	RS485_IN		= 1;

const byte /* Miscellaneous constants used by this module */
	SIGNED	= 0x80,
	UNSIGNED	= 0x00;   /* old 0x00 0x80 */

/* Global variables within this module */
static word  TimeOut = 1000;
static int 	 ErrCode	= 0;
static byte  DevAddr	= 0;
static word	/* base port address of the UARTs Com1 to Com5 */
	COM_BASE [5] = { 0x3F8, 0x2F8, 0x3E8, 0x2E8, 0x3F8 };
static word	 ComBase = 0x3F8;		/* Current Com No. base port address */
static reg32 DataBuf;

double round (double N)
{
	if ((N - floor (N)) < 0.5)
		return floor (N);
	else
		return ceil (N);
}

#define	SetBit8()		outportb (ComBase + LCR, WLEN | PEN | STICK)
#define	ClrBit8()		outportb (ComBase + LCR, WLEN | EVEN | PEN | STICK)
#define	BufferEmpty()	((inportb (ComBase + LSR) & THRE) == THRE)
#define	ShiftEmpty()	((inportb (ComBase + LSR) & TSRE) == TSRE)
#define	DataReady()		((inportb (ComBase + LSR) & DR) == DR)
#define	BrkDetect()		((inportb (ComBase + LSR) & BI) == BI)
#define	OutputRS485()	outportb (ComBase + RS485_REG, RS485_OUT)
#define	InputRS485()	outportb (ComBase + RS485_REG, RS485_IN)
#define	WriteBuf(D)		outportb (ComBase + THR, D)
#define	ReadBuf()		(inportb (ComBase + RBR))

/*****************************************************************************
	Function:
		Initializes the ComNo to the given BaudRate and the other parameters
		needed by the Jenco Serial Communications Protocol
	Input Parameters:
		ComNo 	= 1 to 5
		BaudRate = 300, 600, 1200, 2400, 4800, 9600
   Returns: none

   Code :
		= NO_ERROR if initialization was successful,
		= INVALID_COM if the ComNo is not supported,
		= INVALID_BAUD if the BaudRate is not supported.
		= NO_COM if there is no serial port at the given ComNo
	Warning:
		This must be called first before any of the other routines is called
		in this module, otherwise unpredictable behavior will occur.
*****************************************************************************/
void InitCom (byte ComNo, word BaudRate, int *Code)
{
	reg16 divisor;

	SwitchToCom(ComNo,Code);
	if (*Code != NO_ERROR)
		{ *Code = ErrCode;
			return;
		}
	switch (BaudRate) {
		case 300:	divisor.Word = 384; break;
		case 600:	divisor.Word = 192; break;
		case 1200:	divisor.Word = 96; break;
		case 2400:	divisor.Word = 48; break;
		case 4800:	divisor.Word = 24; break;
		case 9600:	divisor.Word = 12; break;
		default:    *Code = INVALID_BAUD;
						return;
	}
	outportb (ComBase + LCR, DLAB | STICK | EVEN | PEN | WLEN );
	outportb (ComBase + DLL, divisor.Byte [0]);
	outportb (ComBase + DLM, divisor.Byte [1]);
	outportb (ComBase + LCR, STICK | EVEN | PEN | WLEN);
	outportb (ComBase + IER, 0x00 /*| ERLS*/);	/* no interrupts from this com! */
	/* perform the loopback test to see if the Com really exists */
	outportb (ComBase + MCR, LOOP);
	WriteBuf (0xAA);
	delay (40);					/* this is to compensate for the slowest speed */
	if (ReadBuf () != 0xAA)	/* at 300 baud, after writing the receive buffer */
		{ *Code = NO_COM;
		  return;
		}                    /* does not immediately get filled */
	WriteBuf (0x55);
	delay (40);
	if (ReadBuf () != 0x55)
		{ *Code= NO_COM;
		  return;
		}
	outportb (ComBase + MCR, DTR);	/* DTR = +12, RTS = -12; */
	*Code = NO_ERROR;
}

/*****************************************************************************
	Function:
		Sets the base address for Com5
	Input Parameters:
		PortAddr = 0x200 to 0x3F8
	
   Returns: none
	
   Code :	
      = NO_ERROR if Com5 set properly
		= BAD_USER_COM if the Addr is not within 0x200 to 0x3F8
	Warning:
		The Addr is limited to this range, since anything below 0x200 is
		used by the main system board, while 0x3F8 is the highest address
		available for most serial cards, any port address within this range
		that conflicts with other cards may cause damage to the computer
		or its peripherals. The user is cautioned to use discretion in
		selecting an unused port address for the serial card.
*****************************************************************************/
void SetUserCom (word PortAddr, int *Code)
{
	if (PortAddr < 0x200 || PortAddr > 0x3F8)
		{ *Code = BAD_USER_COM;
		  return ;
		}
	else
		{ COM_BASE [4] = PortAddr;
		  *Code = NO_ERROR;
		  return;
		}
}

/*****************************************************************************
	Input Parameters:
		ComNo = 1 to 5
 
   Returns: none
 
   Code :
		= NO_ERROR if ComNo is within 1 to 5
		= INVALID_COM if the ComNo is not supported.
*****************************************************************************/
void SwitchToCom (byte ComNo, int *Code)
{
	if (ComNo < 1 || ComNo > 5)
		{ *Code = INVALID_COM;
			return;
		}
	else
	  { ComBase = COM_BASE [ComNo-1];
		 *Code = NO_ERROR;
		 return;
	  }
}

/*****************************************************************************
	Input Parameters:
		Address = device address from 0 to 255
	Returns:
		none
*****************************************************************************/
void SwitchToAddress (byte Address)
{
	DevAddr = Address;
}

/*****************************************************************************
	Input Parameters:
		Data = data outputted to the com port
	Returns:
		none
*****************************************************************************/
void SendCom (byte Data)
{
   while (!BufferEmpty ());
	OutputRS485 ();
   delay(1);
	WriteBuf (Data);
   while (!ShiftEmpty ());
	InputRS485 ();
}

/*****************************************************************************
	Input Parameters:
		none
	Returns:
		Data read from the com port if successful, 0 if not
	Notes:
		Sets ErrCode to TIMED_OUT if the device did not respond within a
		certain time limit
*****************************************************************************/
byte ReadCom (void)
{
	int countdown = TimeOut;

	ErrCode = NO_ERROR;
	while (!DataReady ()) {
		countdown--;
		if (countdown == 0) {
			ErrCode = TIMED_OUT;
			return 0;
		}
		delay (1);
	}
	return ReadBuf ();
}

/*****************************************************************************
	Input Parameters:
		Data = byte to be sent to the device
	Returns:
		Data read from the device if the device responded
	Notes:
		Sets ErrCode to TIMED_OUT if the device did not respond within a
		certain time limit, or to DEVICE_BUSY if the device is currently busy
*****************************************************************************/
byte Dialogue (byte Data)
{
	byte reply;

	SendCom (Data);
	reply = ReadCom ();
	if (ErrCode != NO_ERROR) {
		SetBit8 (); 
      SendCom (DevAddr);
      ClrBit8 ();
		ReadCom ();
		if (ErrCode != TIMED_OUT)
			ErrCode = DEVICE_BUSY;
	}
	return reply;
}

/*****************************************************************************
	Input Parameters:
		ParamNo = parameter number to read from
		Data = location to store the read parameter, 0 if not successful
		Code = indicates whether the read was successful,
				 NO_ERROR means Data has been load properly
				 positive values indicate a message from the device
				 negative values indicates a communications error of some sort,
				 possible errors are:
				 TIMED_OUT, DEVICE_BUSY, UNKNOWN_PARAM,
				 UNKNOWN_CMD and INVALID_DATA
	Returns:
		none
*****************************************************************************/
void ReadData (word ParamNo, double *Data, int *Code)
{
	reg16 	param_no;
	int		length, i;
	byte		reply, filler;
	int		exponent;
	float		multiplier, result;

	*Data = 0; 
   *Code = NO_ERROR;
	param_no.Word = ParamNo;
	SetBit8 ();
   reply = Dialogue (DevAddr);
   ClrBit8 ();
	switch (ErrCode) {
		case TIMED_OUT:
		case DEVICE_BUSY:
		*Code = ErrCode;
		return;
	}
/*	if (reply != param_no.Byte [1]) {
		*Code = ErrCode = UNKNOWN_PARAM;
		return;
	}   */
	reply = Dialogue (READ_DATA);
	if (ErrCode != NO_ERROR) {
		*Code = ErrCode;
		return;
	}
	if (reply == NACK) {
		*Code = ErrCode = UNKNOWN_CMD;
		return;
	}
	reply = Dialogue (param_no.Byte [0]);
	if (ErrCode != NO_ERROR) {
		*Code = ErrCode;
		return;
	}
	switch (reply) {
		case MSGS:	*Code = ErrCode = INVALID_DATA; return;
		case BUSY:	*Code = ErrCode = DEVICE_BUSY; return;
		case NACK:	*Code = ErrCode = UNKNOWN_PARAM; return;
		case MSCD:
			reply = Dialogue (RTR);
			*Code = (ErrCode == NO_ERROR ? reply : ErrCode);
			return;
		default:
			length = reply & 0x03;
			DataBuf.Dword = 0;
			for (i = length; i >= 0; i--) {
				DataBuf.Byte [i] = Dialogue (RTR);
				if (ErrCode != NO_ERROR) {
					*Code = ErrCode;
					return;
				}
			}
			filler = 0;
			if ((reply & 0x80) == SIGNED && (DataBuf.Byte [length] & 0x80) == 0x80)
				filler = 0xFF;		/* if signed and negative */
			for (i = length+1; i < 4; i++) DataBuf.Byte [i] = filler;
			exponent = (reply >> 2) & 0x1F;
			if ((exponent & 0x10) == 0x10) exponent |= 0xFFE0;
			multiplier = pow10 (exponent);
			if (filler == 0xFF)
				result = DataBuf.Long;
			else
				result = DataBuf.Dword;
			result *= multiplier;
	}
	*Data = result;
	return;
}

/*****************************************************************************
	Input Parameters:
		ParamNo = parameter number to read from
		Data = data to write to the ParamNo location
		Code = indicates whether the write was successful,
				 NO_ERROR means Data has been load properly
				 negative values indicates a communications error of some sort,
				 positive values indicate a message from the device,
				 possible values are:
				 TIMED_OUT, DEVICE_BUSY, UNKNOWN_PARAM,
				 UNKNOWN_CMD and INVALID_DATA
	Returns:
		none
*****************************************************************************/
void WriteData (word ParamNo, double Data, int *Code)
{
	reg16 	param_no;
	byte		length, sign, reply, i;
	int		exponent;
	double	divisor, temp, lower, upper;
	reg32		result;

	*Code = NO_ERROR;
	param_no.Word = ParamNo;
	SetBit8 (); reply = Dialogue (DevAddr); ClrBit8 (); 
	switch (ErrCode) {
		case TIMED_OUT:
		case DEVICE_BUSY:	*Code = ErrCode; return;
	}
	if (reply != param_no.Byte [1]) {
		*Code = ErrCode = UNKNOWN_PARAM;
		return;
	}
	reply = Dialogue (WRITE_DATA);
	if (ErrCode != NO_ERROR) {
		*Code = ErrCode;
		return;
	}
	if (reply == NACK) {
		*Code = ErrCode = UNKNOWN_CMD;
		return;
	}
	reply = Dialogue (param_no.Byte [0]);
	if (ErrCode != NO_ERROR) {
		*Code = ErrCode;
		return;
	}
	switch (reply) {
		case MSGS:	*Code = ErrCode = INVALID_DATA; return;
		case BUSY:	*Code = ErrCode = DEVICE_BUSY; return;
		case NACK:	*Code = ErrCode = UNKNOWN_PARAM; return;
		case MSCD:
			reply = Dialogue (RTR);
			*Code = ErrCode == NO_ERROR ? reply : ErrCode;
			return;
		default:
			length = reply & 0x03;
			sign = reply & 0x80;
			exponent = (reply >> 2) & 0x1F;
			if ((exponent & 0x10) == 0x10) exponent |= 0xFFE0;
			divisor = pow10 (exponent);
			temp = round (Data / divisor);
			lower = 0.0;
			upper = 1.0;
			for (i = 0; i <= length; i++) {
				upper *= 256.0;
			}
			if (sign == SIGNED) {
				upper = upper / 2;
				lower = -upper;
				if (temp < lower || temp >= upper) {
					*Code = ErrCode = INVALID_DATA;
					return;
				} else {
					result.Long = (long) temp;
				}
			} else {
				if (temp < lower || temp >= upper) {
					*Code = ErrCode = INVALID_DATA;
					return;
				} else {
					result.Dword = (dword) temp;
				}
			}
	}
	for (i = 0; i <= length; i++) {
		reply = Dialogue (result.Byte [i]);
		if (ErrCode != NO_ERROR) {
			*Code = ErrCode;
			return;
		}
		if (reply == NACK) {
			*Code = ErrCode = INVALID_DATA;
			return;
		}
	}
}

/*****************************************************************************
	Input Parameters:
		ParamNo = parameter number of the flag
		BitNo = bit position
		Flag = location to put the read flag
				 0 if the flag is low, 1 if the flag is high
		Data = data to write to the ParamNo location
		Code = indicates whether the write was successful,
				 NO_ERROR means Data has been load properly
				 positive values indicate a message from the device
				 negative values indicates a communications error of some sort,
				 possible values are:
				 TIMED_OUT, DEVICE_BUSY, UNKNOWN_PARAM,
				 UNKNOWN_CMD and INVALID_DATA
	Returns:
		none
*****************************************************************************/
void ReadFlag (word ParamNo, byte BitNo, int *Flag, int *Code)
{
	double dummy;

	*Flag = 0; 
	ReadData (ParamNo, &dummy, Code);
	if (*Code == NO_ERROR)
		*Flag = (int) ((DataBuf.Dword >> BitNo) & 0x1);
	else
		*Flag = -1;
}

/*****************************************************************************
	Input Parameters:
		ParamNo = parameter number of the flag
		Flags = location to put the read flags
				  each bit indicates the state of the particular flag
		Code = indicates whether the write was successful,
				 NO_ERROR means Data has been load properly
				 positive values indicate a message from the device
				 negative values indicates a communications error of some sort,
				 possible values are:
				 TIMED_OUT, DEVICE_BUSY, UNKNOWN_PARAM,
				 UNKNOWN_CMD and INVALID_DATA
	Returns:
		none
*****************************************************************************/
void ReadFlags (word ParamNo, dword *Flags, int *Code)
{
	double dummy;

	*Flags = 0;
	ReadData (ParamNo, &dummy, Code);
	if (*Code == NO_ERROR) 
		*Flags = DataBuf.Dword;
}

/*****************************************************************************
	Input Parameters: none

	Returns: none
   
   Dev_type : The device type at the current device address
*****************************************************************************/
void GetDevType (word *dev_type, int *Code)
{
	byte trycmd = 0, reply;

	do {
		SetBit8 ();
		*dev_type = Dialogue (DevAddr) << 8;
		ClrBit8 ();
		if (ErrCode != NO_ERROR)
			{ *Code = ErrCode;
			  return;
			}
		reply = Dialogue (trycmd);
		if (reply != NACK)
			{ *dev_type = *dev_type | reply;
			  return;
			}
		if (++trycmd > 5)
			{ *Code = UNKNOWN_CMD;
			  return;
			}
	} while (1);
}

#define	SPECIAL_CMDS	8
#define	OFFSET			16

/*****************************************************************************
	Input Parameters:
		Cmd - the command specific to the given device

	Returns : none

   Code :
		= NO_ERROR if the command was executed successfully
		= UNKNOWN_CMD if the command is unrecognized and ignored by the device
		= TIMED_OUT if the device did not respond within a certain period
		= DEVICE_BUSY if the device cannot process the command immediately
*****************************************************************************/
void Invoke (int Cmd, int *Code)
{
	byte reply;

	SetBit8 (); reply = Dialogue (DevAddr); ClrBit8 ();

	if (Cmd < SPECIAL_CMDS) {	/* 0 to 7 are special commands */
		reply = Dialogue (Cmd + OFFSET);	/* offset for the first 16 reserved */
		if (ErrCode != NO_ERROR)
			{ *Code = ErrCode;
			  return;
			}
		if (reply == NACK)
			{ *Code =UNKNOWN_CMD;
			  return;
			}
	} else
		  {
			 reply = Dialogue (INVOKE_CMD);
			 if (ErrCode != NO_ERROR)
				 { *Code =ErrCode;
					return;
				 }
			if (reply == NACK)
				 { *Code =UNKNOWN_CMD;
					return;
				 }
		reply = Dialogue (Cmd - SPECIAL_CMDS);
		if (ErrCode != NO_ERROR)
			{ *Code = ErrCode;
			  return;
			}
		if (reply == NACK)
			{ *Code =UNKNOWN_PARAM;
			  return;
			}
	}
	*Code = NO_ERROR;
}

